/*
 * extract.c -- extract a file from simulated disk
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define BLKSIZE		512
#define USR_START	(0x8FB000 / BLKSIZE)

#define NICFREE		50	/* number of superblock free blocks */
#define NICINOD		100	/* number of superblock inodes */

#define OK		0
#define ERROR		1


typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned long U32;
typedef signed long S32;


typedef S32		my_daddr_t;
typedef U16		my_ino_t;
typedef S32		my_time_t;
typedef S32		my_off_t;


typedef struct  filsys {
  U16            s_isize;          /* size in blocks of i-list */
  my_daddr_t     s_fsize;          /* size in blocks of entire volume */
  S16            s_nfree;          /* number of addresses in s_free */
  my_daddr_t     s_free[NICFREE];  /* free block list */
  S16            s_ninode;         /* number of i-nodes in s_inode */
  my_ino_t       s_inode[NICINOD]; /* free i-node list */
  S8             s_flock;          /* lock during free list manipulation */
  S8             s_ilock;          /* lock during i-list manipulation */
  S8             s_fmod;           /* super block modified flag */
  S8             s_ronly;          /* mounted read-only flag */
  my_time_t      s_time;           /* last super block update */
  /* remainder not maintained by this version of the system */
  my_daddr_t     s_tfree;          /* total free blocks*/
  my_ino_t       s_tinode;         /* total free inodes */
  S16            s_m;              /* interleave factor */
  S16            s_n;              /* " " */
  S8             s_fname[6];       /* file system name */
  S8             s_fpack[6];       /* file system pack name */
} __attribute__((packed)) SB;


typedef struct dinode {
  U16            di_mode;          /* mode and type of file */
  S16            di_nlink;         /* number of links to file */
  S16            di_uid;           /* owner's user id */
  S16            di_gid;           /* owner's group id */
  my_off_t       di_size;          /* number of bytes in file */
  S8             di_addr[40];      /* disk block addresses */
  my_time_t      di_atime;         /* time last accessed */
  my_time_t      di_mtime;         /* time last modified */
  my_time_t      di_ctime;         /* time created */
} __attribute__((packed)) INO;


#define INOPB		8	/* 8 inodes per block */
#define ROOT_INO	2	/* root inode */


FILE *disk;
int blocks;
U8 bootBlock[BLKSIZE];
U8 superBlock[BLKSIZE];
U8 inodeBlock[BLKSIZE];
U8 singleIndirectBlock[BLKSIZE];
U8 *data;


void error(char *msg) {
  printf("Error: %s\n", msg);
  exit(1);
}


U16 readShort(U8 *buf) {
  return (U16) buf[0] | ((U16) buf[1] << 8);
}


U32 readLong(U8 *buf) {
  return ((U32) buf[0] << 16) | ((U32) buf[1] << 24) |
         (U32) buf[2] | ((U32) buf[3] << 8);
}


void showBlock(U8 *buffer) {
  int i, j;
  U8 c;

  for (i = 0; i < BLKSIZE/16; i++) {
    printf("%04X   ", i * 16);
    for (j = 0; j < 16; j++) {
      c = buffer[i * 16 + j];
      printf("%02X ", c);
    }
    printf("   ");
    for (j = 0; j < 16; j++) {
      c = buffer[i * 16 + j];
      if (c >= 0x20 && c < 0x7F) {
        printf("%c", c);
      } else {
        printf(".");
      }
    }
    printf("\n");
  }
}


int readBlock(int block, U8 *buffer) {
  block += USR_START;
  if (fseek(disk, (long) block * BLKSIZE, SEEK_SET) != 0) {
    return ERROR;
  }
  if (fread(buffer, 1, BLKSIZE, disk) != BLKSIZE) {
    return ERROR;
  }
  return OK;
}


INO *readInode(my_ino_t inodeNumber) {
  int block;
  int offset;

  block = (inodeNumber + 15) / INOPB;
  offset = ((inodeNumber + 15) % INOPB) * sizeof(INO);
  if (readBlock(block, inodeBlock) != OK) {
    error("cannot read inode");
  }
  return (INO *) &inodeBlock[offset];
}


void readDataBlock(INO *inop, int blkno, U8 *buffer) {
  U8 a, b, c, d;
  int blockNo;

  if (blkno < 10) {
    /* direct block */
    a = inop->di_addr[blkno * 3 + 0];
    b = inop->di_addr[blkno * 3 + 1];
    c = inop->di_addr[blkno * 3 + 2];
    printf("addr %02X %02x %02X\n", a, b, c);
    blockNo = ((U32) a << 24) | ((U32) b) | ((U32) c << 8);
    if (readBlock(blockNo, buffer) != OK) {
      error("cannot read data block");
    }
  } else
  if (blkno < 10 + BLKSIZE / 4) {
    /* single indirect block */
    a = inop->di_addr[10 * 3 + 0];
    b = inop->di_addr[10 * 3 + 1];
    c = inop->di_addr[10 * 3 + 2];
    printf("addr %02X %02x %02X\n", a, b, c);
    blockNo = ((U32) a << 24) | ((U32) b) | ((U32) c << 8);
    if (readBlock(blockNo, singleIndirectBlock) != OK) {
      error("cannot read single indirect block");
    }
    printf("Single Indirect Block:\n");
    showBlock(singleIndirectBlock);
    printf("\n");
    a = singleIndirectBlock[(blkno - 10) * 4 + 0];
    b = singleIndirectBlock[(blkno - 10) * 4 + 1];
    c = singleIndirectBlock[(blkno - 10) * 4 + 2];
    d = singleIndirectBlock[(blkno - 10) * 4 + 3];
    printf("addr %02X %02x %02X %02X\n", a, b, c, d);
    blockNo = ((U32) a << 16) | ((U32) b << 24) |
              ((U32) c) | ((U32) d << 8);
    if (readBlock(blockNo, buffer) != OK) {
      error("cannot read data block");
    }
  } else {
    error("double/triple indirect blocks not yet");
  }
}


int main(int argc, char *argv[]) {
  U32 size;
  SB *sbp;
  U16 s_isize;
  U32 s_fsize;
  U16 s_nfree;
  U16 s_ninode;
  int ino;
  char *path;
  INO *inop;
  U16 di_nlink;
  U32 di_size;
  int dataBlocks;
  int bytesInLastBlock;
  int i;
  int d;
  char *p1, *p2;
  int found;
  FILE *dst;

  if (argc != 4) {
    printf("Usage: %s <disk image> <src file relative to /usr> <dst file>\n",
           argv[0]);
    exit(1);
  }
  disk = fopen(argv[1], "rb");
  if (disk == NULL) {
    error("cannot open disk");
  }
  fseek(disk, 0, SEEK_END);
  size = ftell(disk);
  if (size % BLKSIZE) {
    error("disk not multiple of blocksize");
  }
  blocks = size / BLKSIZE;
  printf("Disk has %d blocks\n\n", blocks);
  if (readBlock(0, bootBlock) != OK) {
    error("cannot read boot block");
  }
  printf("Boot Block:\n");
  showBlock(bootBlock);
  printf("\n");
  if (readBlock(1, superBlock) != OK) {
    error("cannot read super block");
  }
  printf("Super Block:\n");
  showBlock(superBlock);
  printf("\n");
  sbp = (SB *) superBlock;
  s_isize = readShort((U8 *) &sbp->s_isize);
  printf("size of i-list in blocks          %hu\n", s_isize);
  s_fsize = readLong((U8 *) &sbp->s_fsize);
  printf("size of entire volume in blocks   %lu\n", s_fsize);
  s_nfree = readShort((U8 *) &sbp->s_nfree);
  printf("number of addresses in s_free     %hu\n", s_nfree);
  s_ninode = readShort((U8 *) &sbp->s_ninode);
  printf("number of i-nodes in s_inode      %hu\n", s_ninode);
  ino = ROOT_INO;
  path = argv[2];
  if (*path == '/') {
    path++;
  }
  while (1) {
    inop = readInode(ino);
    printf("Inode Block:\n");
    showBlock(inodeBlock);
    printf("\n");
    di_nlink = readShort((U8*) &inop->di_nlink);
    printf("this file has %hu links\n", di_nlink);
    di_size = readLong((U8*) &inop->di_size);
    printf("this file has size %lu bytes\n", di_size);
    dataBlocks = di_size / BLKSIZE;
    bytesInLastBlock = di_size % BLKSIZE;
    if (bytesInLastBlock != 0) {
      dataBlocks++;
    }
    data = malloc((long) dataBlocks * BLKSIZE);
    if (data == NULL) {
      error("out of memory");
    }
    for (i = 0; i < dataBlocks; i++) {
      readDataBlock(inop, i, data + i * BLKSIZE);
      printf("Data Block %d:\n", i);
      showBlock(data + i * BLKSIZE);
      printf("\n");
    }
    if (*path == '\0') {
      break;
    }
    for (d = 0; d < di_size/16; d++) {
      ino = readShort(data + d * 16);
      if (ino == 0) {
        continue;
      }
      p1 = data + d * 16 + 2;
      p2 = path;
      found = 0;
      while (1) {
        if (*p2 != *p1) {
          break;
        }
        p1++;
        p2++;
        if (*p2 == '/' || *p2 == '\0') {
          if (*p1 == '\0') {
            found = 1;
          }
          break;
        }
      }
      if (found) {
        path = p2;
        if (*path == '/') {
          path++;
        }
        break;
      }
    }
    if (d == di_size/16) {
      error("file not found");
    }
  }
  fclose(disk);
  dst = fopen(argv[3], "wb");
  if (dst == NULL) {
    error("cannot open destination file");
  }
  if (fwrite(data, 1, di_size, dst) != di_size) {
    error("canot write destination file");
  }
  fclose(dst);
  return 0;
}
